// proc.C
//
// This program is free software. See the file COPYING for details.
// Author: Mattias Engdegrd, 1997, 1998

#include <stdio.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <dirent.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sched.h>

#include "qps.h"
#include "proc.h"
#include "svec.C"
#include "uidstr.h"
#include "ttystr.h"
#include "wchan.h"
#include "details.h"

//#define FAKE_SMP 4	// for SMP debugging on UP machines

// socket states, from <linux/net.h> and touched to avoid name collisions
enum {
  SSFREE = 0,			/* not allocated		*/
  SSUNCONNECTED,		/* unconnected to any socket	*/
  SSCONNECTING,			/* in process of connecting	*/
  SSCONNECTED,			/* connected to socket		*/
  SSDISCONNECTING		/* in process of disconnecting	*/
};

char *procdir = "/proc";

int Procinfo::page_k_shift;

Procinfo::Procinfo(int proc_pid)
{
    if(readproc(proc_pid) < 0)
	pid = -1;		// invalidate object, will be deleted
    selected = FALSE;
    viewed = FALSE;
    details = 0;
    sock_inodes = 0;
    fd_files = 0;
    maps = 0;
    environ = 0;
    envblock = 0;
    socks_current = FALSE;
    usocks_current = FALSE;
}

Procinfo::~Procinfo()
{
    delete sock_inodes;
    if(maps) {
	maps->purge();
	delete maps;
    }
    if(fd_files) {
	fd_files->purge();
	delete fd_files;
    }
    delete environ;
    if(envblock)
	free(envblock);
}

// miscellaneous static initializations
void Procinfo::init_static()
{
    if(num_cpus == 0) {
	num_cpus = count_cpus();
	cpu_times_vec = new unsigned[CPUTIMES * num_cpus];
	old_cpu_times_vec = new unsigned[CPUTIMES * num_cpus];
    }

    socks.setAutoDelete(TRUE);
    usocks.setAutoDelete(TRUE);

    page_k_shift = 0;
    for(int j = getpagesize(); j > 1024; j >>= 1)
	page_k_shift++;
}

// return number of bytes read if ok, -1 if failed
int Procinfo::read_file(char *name, char *buf, int max)
{	
    int fd = open(name, O_RDONLY);
    if(fd < 0) return -1;
    int r = read(fd, buf, max);
    if(r < 0) return -1;
    close(fd);
    return r;
}

int Procinfo::readproc(int proc_pid)
{
    char path[256];
    char buf[256];
    char sbuf[4096];		// should be enough to acommodate /proc/X/stat
    char cmdbuf[MAX_CMD_LEN];

    sprintf(path, "%s/%d", procdir, proc_pid);
    pid = proc_pid;

    // read /proc/XX/cmdline
    strcpy(buf, path);
    strcat(buf, "/cmdline");
    int cmdlen;
    if((cmdlen = read_file(buf, cmdbuf, MAX_CMD_LEN - 1)) < 0) return -1;
    if(cmdlen == 0) {
	cmdline = "";
    } else {
	for(int i = 0; i < cmdlen; i++)
	    if(cmdbuf[i] == '\0') cmdbuf[i] = ' ';
	int j = cmdlen - 1;
	while(j >= 0 && cmdbuf[j] == ' ')
	    j--;
	cmdbuf[j + 1] = '\0';
	cmdline = cmdbuf;
    }

    // read /proc/XX/stat
    strcpy(buf, path);
    strcat(buf, "/stat");
    int statlen;
    if((statlen = read_file(buf, sbuf, sizeof(sbuf) - 1)) <= 0) return -1;
    sbuf[statlen] = '\0';
    char *p = strrchr(sbuf, ')');
    *p = '\0';			// split in two parts
    comm = strchr(sbuf, '(') + 1;

    // dummy variables: not necessary but nice to remember which value goes
    // where
    unsigned long itrealvalue, vsize, startcode, endcode, startstack,
	          kstkesp, kstkeip, signal, blocked, sigignore, sigcatch,
	          rlim;
    sscanf(p + 2, "%c %d %d %d %d %d %lu %lu %lu %lu %lu "
	   "%ld %ld %ld %ld %d %d %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu "
	   "%lu %lu %lu %lu %lu",
	   &state, &ppid, &pgrp, &session, &tty, &tpgid,
	   &flags, &minflt, &cminflt, &majflt, &cmajflt,
	   &utime, &stime, &cutime, &cstime, &priority, &nice,
	   &timeout, &itrealvalue, &starttime, &vsize, &rss, &rlim,
	   &startcode, &endcode, &startstack, &kstkesp, &kstkeip,
	   &signal, &blocked, &sigignore, &sigcatch, &wchan);
    rss <<= page_k_shift;	// rss given in pages, display in Kbytes
    pmem = 100.0 * rss / mem_total;

    // read /proc/XX/statm
    strcpy(buf, path);
    strcat(buf, "/statm");
    if((statlen = read_file(buf, sbuf, sizeof(sbuf) - 1)) <= 0) return -1;
    sbuf[statlen] = '\0';
    sscanf(sbuf, "%lu %lu %lu %lu %lu %lu %lu",
	   &size, &resident, &share, &trs, &lrs, &drs, &dt);
    size <<= page_k_shift;
    resident <<= page_k_shift;
    share <<= page_k_shift;
    trs <<= page_k_shift;
    lrs <<= page_k_shift;
    drs <<= page_k_shift;

    // read /proc/XX/status
    strcpy(buf, path);
    strcat(buf, "/status");
    if((statlen = read_file(buf, sbuf, sizeof(sbuf) - 1)) <= 0) return -1;
    sbuf[statlen] = '\0';
    p = sbuf;
    for(int i = 0; i < 4; i++)
	p = strchr(p + 1, '\n');	// skip four lines
    sscanf(p + 1, "Uid: %d %d %d %d Gid: %d %d %d %d",
	   &uid, &euid, &suid, &fsuid,
	   &gid, &egid, &sgid, &fsgid);

    gettimeofday(&tv, 0);
    policy = -1;		// will get it when needed
    rtprio = -1;		// ditto

    return pid;
}

float Procinfo::loadavg[] = {0.0, 0.0, 0.0};
int Procinfo::mem_total = 0;
int Procinfo::mem_free = 0;
int Procinfo::mem_shared = 0;
int Procinfo::mem_buffers = 0;
int Procinfo::mem_cached = 0;
int Procinfo::swap_total = 0;
int Procinfo::swap_free = 0;
unsigned *Procinfo::cpu_times_vec = 0;
unsigned *Procinfo::old_cpu_times_vec = 0;
unsigned Procinfo::boot_time = 0;
int Procinfo::uptime = 0;
QIntDict<Sockinfo> Procinfo::socks(73);
bool Procinfo::socks_current = FALSE;
QIntDict<UnixSocket> Procinfo::usocks(17);
bool Procinfo::usocks_current = FALSE;
unsigned int Procinfo::num_cpus = 0;


// just grab the load averages
void Procinfo::read_loadavg()
{
    char path[80];
    char buf[512];
    strcpy(path, procdir);
    strcat(path, "/loadavg");
    int n;
    if((n = read_file(path, buf, sizeof(buf) - 1)) <= 0) {
	fprintf(stderr,	"qps: Cannot open /proc/loadavg"
		        " (make sure /proc is mounted)\n");
	exit(1);
    }
    buf[n] = '\0';
    sscanf(buf, "%f %f %f", &loadavg[0], &loadavg[1], &loadavg[2]);
}

// read information common to all processes
void Procinfo::read_common()
{
    char path[80];
    char buf[4096 + 1];

    // read memory info
    strcpy(path, procdir);
    strcat(path, "/meminfo");
    int n;
    if((n = read_file(path, buf, sizeof(buf) - 1)) <= 0) return;
    buf[n] = '\0';

    // Skip the old /meminfo cruft, making this work in post-2.1.42 kernels
    // as well.  (values are now in kB)
    char *p = strstr(buf, "MemTotal:");
    sscanf(p, "MemTotal: %d kB\nMemFree: %d kB\nMemShared: %d kB\nBuffers: %d"
              " kB\nCached: %d kB\nSwapTotal: %d kB\n SwapFree: %d",
	   &mem_total, &mem_free, &mem_shared, &mem_buffers,
	   &mem_cached, &swap_total, &swap_free);

    // read system status
    strcpy(path, procdir);
    strcat(path, "/stat");
    if((n = read_file(path, buf, sizeof(buf) - 1)) <= 0) return;
    buf[n] = '\0';

    for(unsigned cpu = 0; cpu < num_cpus; cpu++)
	for(int i = 0; i < CPUTIMES; i++)
	    old_cpu_times(cpu, i) = cpu_times(cpu, i);

    if(num_cpus == 1) {
	sscanf(buf, "cpu %u %u %u %u",
	       &cpu_times(0, CPU_USER), &cpu_times(0, CPU_NICE),
	       &cpu_times(0, CPU_SYSTEM), &cpu_times(0, CPU_IDLE));
    } else {
#ifdef FAKE_SMP
	sscanf(buf, "cpu %u %u %u %u",
	       &cpu_times(0, CPU_USER), &cpu_times(0, CPU_NICE),
	       &cpu_times(0, CPU_SYSTEM), &cpu_times(0, CPU_IDLE));
	for(unsigned cpu = 1; cpu < num_cpus; cpu++) {
	    for(int i = 0; i < CPUTIMES; i++)
		cpu_times(cpu, i) = cpu_times(0, i);
	}
#else
	// SMP
	for(unsigned cpu = 0; cpu < num_cpus; cpu++) {
	    char cpu_buf[10];
	    sprintf(cpu_buf, "cpu%d", cpu);
	    if((p = strstr(buf, cpu_buf)) != 0) {
		sscanf(p, "%*s %u %u %u %u",
		       &cpu_times(cpu, CPU_USER), 
		       &cpu_times(cpu, CPU_NICE),
		       &cpu_times(cpu, CPU_SYSTEM), 
		       &cpu_times(cpu, CPU_IDLE));
	    } else {
		printf("Error reading info for cpu %d\n", cpu);
		abort();
	    }
	}
#endif
    }

    // 2.0.x kernels (at least up to 2.0.33) have an SMP bug that reports
    // cpu_time(CPU_IDLE) incorrectly, since it doesn't take the number of
    // cpus into account. This is fixed in 2.1.x kernels, but a workaround
    // should really be here.

    p = strstr(buf, "btime") + 6;
    sscanf(p, "%u", &boot_time);

    // read uptime
    strcpy(path, procdir);
    strcat(path, "/uptime");
    if((n = read_file(path, buf, sizeof(buf) - 1)) < 0) return;
    buf[n] = '\0';
    float up_s;
    sscanf(buf, "%f", &up_s);
    uptime = (int)up_s;
}

unsigned int Procinfo::count_cpus()
{
#ifdef FAKE_SMP
    return FAKE_SMP;
#else
    unsigned num_cpus = 0;
    char path[512], buf[4096];
    char *p;

    strcpy(path, procdir);
    strcat(path, "/stat");
    int n;
    if((n = read_file(path, buf, sizeof(buf) - 1)) <= 0)
	return 1;
    buf[n] = '\0';

    if((p = strstr(buf, "cpu")) != 0) {
	while(p < buf + sizeof(buf) - 4 && strncmp(p, "cpu", 3) == 0) {
	    num_cpus++;
	    if(strncmp(p, "cpu0", 4) == 0)
		num_cpus--;
	    p = strchr(p, '\n');
	    if (p != NULL)
		p++;
	}
    }
    return num_cpus;
#endif
}

int Procinfo::get_policy()
{
    if(policy == -1)
	policy = sched_getscheduler(pid);
    return policy;
}

int Procinfo::get_rtprio()
{
    if(rtprio < 0) {
	struct sched_param p;
	if(sched_getparam(pid, &p) == 0)
	    rtprio = p.sched_priority;
    }
    return rtprio;
}

// return TRUE if /proc/PID/fd could be read, FALSE otherwise
// store fileinfo, and also socket inodes separately
bool Procinfo::read_fds()
{
    char path[80], *p;
    sprintf(path, "%s/%d/fd", procdir, pid);

    DIR *d = opendir(path);
    if(!d) return FALSE;

    if(!fd_files)
	fd_files = new Svec<Fileinfo*>;
    fd_files->clear();

    if(!sock_inodes)
	sock_inodes = new Svec<SockInode>;
    sock_inodes->clear();

    p = path + strlen(path) + 1;
    p[-1] = '/';

    struct dirent *e;
    char buf[80];
    while((e = readdir(d)) != 0) {
	if(e->d_name[0] == '.')
	    continue;		// skip . and ..
	strcpy(p, e->d_name);
	int fdnum = atoi(p);
	int len;
	if( (len = readlink(path, buf, sizeof(buf) - 1)) > 0) {
	    buf[len] = '\0';
	    unsigned long dev, ino;
	    if((buf[0] == '['	// Linux 2.0 style /proc/fd
		&& sscanf(buf, "[%lx]:%lu", &dev, &ino) == 2
		&& dev == 0)
	       || sscanf(buf, "socket:[%lu]", &ino) > 0) { // Linux 2.1
		Sockinfo *si = Procinfo::socks[ino];
		if(si) {
		    // a TCP or UDP socket
		    sock_inodes->add(SockInode(fdnum, ino));
		    QString s;
		    s.sprintf("%sp socket %d",
			      si->proto == Sockinfo::TCP ? "tc" : "ud", ino);
		    fd_files->add(new Fileinfo(fdnum, s));
		    continue;
		} else {
		    // maybe a unix domain socket?
		    read_usockets();
		    UnixSocket *us = Procinfo::usocks[ino];
		    if(us) {
			QString s;
			char *tp = "?", *st = "?";
			switch(us->type) {
			case SOCK_STREAM: tp = "stream"; break;
			case SOCK_DGRAM: tp = "dgram"; break;
			}
			switch(us->state) {
			case SSFREE: st = "free"; break;
			case SSUNCONNECTED: st = "unconn"; break;
			case SSCONNECTING: st = "connecting"; break;
			case SSCONNECTED: st = "connected"; break;
			case SSDISCONNECTING: st = "disconn"; break;
			}
			s.sprintf("unix domain socket %d (%s, %s) %s",
				  ino, tp, st, (const char *)us->name);
			fd_files->add(new Fileinfo(fdnum, s));
			continue;
		    }
		}
	    }
	    // assume fds will be read in increasing order
	    fd_files->add(new Fileinfo(fdnum, buf));
	}
    }
    closedir(d);
    return TRUE;
}

bool Procinfo::read_socket_list(Sockinfo::proto_t proto, char *pseudofile)
{
    char path[80];
    strcpy(path, procdir);
    strcat(path, "/net/");
    strcat(path, pseudofile);
    FILE *f = fopen(path, "r");
    if(!f) return FALSE;

    char buf[256];
    fgets(buf, sizeof(buf), f);	// skip header
    while(fgets(buf, sizeof(buf), f) != 0) {
	Sockinfo *si = new Sockinfo;
	si->proto = proto;
	unsigned local_port, rem_port, st, tr;
	sscanf(buf + 6, "%x:%x %x:%x %x %x:%x %x:%x %x %d %d %d",
	       &si->local_addr, &local_port, &si->rem_addr, &rem_port,
	       &st, &si->tx_queue, &si->rx_queue,
	       &tr, &si->tm_when, &si->rexmits,
	       &si->uid, &si->timeout, &si->inode);
	// fix fields that aren't sizeof(int)
	si->local_port = local_port;
	si->rem_port = rem_port;
	si->st = st;
	si->tr = tr;

	socks.insert(si->inode, si);
    }
    fclose(f);
    return TRUE;
}

bool Procinfo::read_usocket_list()
{
    char path[80];
    strcpy(path, procdir);
    strcat(path, "/net/unix");
    FILE *f = fopen(path, "r");
    if(!f) return FALSE;

    char buf[256];
    fgets(buf, sizeof(buf), f);	// skip header
    while(fgets(buf, sizeof(buf), f)) {
	if(buf[0])
	    buf[strlen(buf) - 1] = '\0'; // chomp newline
	UnixSocket *us = new UnixSocket;
	unsigned q;
	unsigned type, state;
	int n;
	sscanf(buf, "%x: %x %x %x %x %x %ld %n",
	       &q, &q, &q, &us->flags, &type, &state, &us->inode, &n);
	us->name = buf + n;
	us->type = type;
	us->state = state;
	usocks.insert(us->inode, us);
    }
    fclose(f);
    return TRUE;
}

void Procinfo::read_sockets()
{
    if(socks_current)
	return;

    socks.clear();
    if(!read_socket_list(Sockinfo::TCP, "tcp")
       || !read_socket_list(Sockinfo::UDP, "udp"))
	return;

    socks_current = TRUE;
}

void Procinfo::read_usockets()
{
    if(usocks_current)
	return;

    usocks.clear();
    if(!read_usocket_list())
	return;

    usocks_current = TRUE;
}

void Procinfo::invalidate_sockets()
{
    socks_current = usocks_current = FALSE;
}

// return TRUE if /proc/XX/maps could be read, FALSE otherwise
bool Procinfo::read_maps()
{
    // idea: here we could readlink /proc/XX/exe to identify the executable
    // when running 2.0.x
    char name[80];
    sprintf(name, "%s/%d/maps", procdir, pid);
    FILE *f = fopen(name, "r");
    if(!f) return FALSE;
    char line[1024];		// lines can be this long, or longer
    if(!maps)
	maps = new Svec<Mapsinfo *>;
    else
	maps->clear();
    
    while(fgets(line, sizeof(line), f)) {
	Mapsinfo *mi = new Mapsinfo;
	int n;
	unsigned int major, minor;
	sscanf(line, "%lx-%lx %4c %lx %x:%x %lu%n",
	       &mi->from, &mi->to, mi->perm, &mi->offset,
	       &major, &minor, &mi->inode, &n);
	mi->major = major; mi->minor = minor;
	if(line[n] != '\n') {
	    int len = strlen(line);
	    if(line[len - 1] == '\n')
		line[len - 1] = '\0';
	    while(line[n] == ' ' && line[n]) n++;
	    mi->filename = line + n;
	} else if((mi->major | mi->minor | mi->inode) == 0)
	    mi->filename = "(anonymous)";
	maps->add(mi);
    }
    fclose(f);
    return TRUE;
}

// return TRUE if /proc/XX/environ could be read, FALSE otherwise
bool Procinfo::read_environ()
{
    if(envblock) free(envblock);
    int bs = 4096;		// good start
    envblock = (char *)malloc(bs + 1);
    char path[128];
    sprintf(path, "%s/%d/environ", procdir, pid);
    int fd = open(path, O_RDONLY);
    if(fd < 0) {
	free(envblock);
	envblock = 0;
	return FALSE;
    }
    int n;
    int ofs = 0;
    while((n = read(fd, envblock + ofs, bs - ofs)) == bs - ofs) {
	ofs = bs;
	envblock = (char *)realloc(envblock, (bs += 4096) + 1);
    }
    close(fd);
    if(n < 0) {
	free(envblock);
	envblock = 0;
	return FALSE;
    }
    n += ofs;
    envblock[n] = '\0';
    if(!environ)
	environ = new Svec<NameValue>(64);
    else
	environ->clear();
    for(int i = 0; i < n;) {
	char *p = strchr(envblock + i, '=');
	if(p)
	    *p++ = '\0';
	else	// degenerate variable: treat as name with empty value
	    p = envblock + i + strlen(envblock + i);
	environ->add(NameValue(envblock + i, p));
	i = p - envblock + strlen(p) + 1;
    }
    return TRUE;
}

int Category::compare(Procinfo *a, Procinfo *b)
{
    return strcmp(string(a), string(b));
}

Cat_int::Cat_int(char *heading, char *explain, int w, int Procinfo::*member)
       : Category(heading, explain), int_member(member), field_width(w)
{}

QString Cat_int::string(Procinfo *p)
{
    QString s(10);
    s.setNum(p->*int_member);	
    return s;
}

int Cat_int::compare(Procinfo *a, Procinfo *b)
{
    // qsort() only cares about the sign of the number returned by the
    // comparison function; only a subtraction is necessary
    return a->*int_member - b->*int_member;
}

Cat_uintl::Cat_uintl(char *heading, char *explain, int w,
		     unsigned long Procinfo::*member)
        : Category(heading, explain), uintl_member(member), field_width(w)
{}

QString Cat_uintl::string(Procinfo *p)
{
    QString s(10);
    s.setNum(p->*uintl_member);	
    return s;
}

int Cat_uintl::compare(Procinfo *a, Procinfo *b)
{
    return b->*uintl_member - a->*uintl_member;
}

Cat_hex::Cat_hex(char *heading, char *explain, int w,
		 unsigned long Procinfo::*member)
        : Cat_uintl(heading, explain, w, member)
{}

QString Cat_hex::string(Procinfo *p)
{
    QString s(11);
    s.sprintf("%8x", p->*uintl_member);
    return s;
}

Cat_swap::Cat_swap(char *heading, char *explain)
        : Category(heading, explain)
{}

QString Cat_swap::string(Procinfo *p)
{
    QString s(10);
    s.setNum(p->size - p->resident);
    return s;
}

int Cat_swap::compare(Procinfo *a, Procinfo *b)
{
    return (b->size - b->resident) - (a->size - a->resident);
}

Cat_string::Cat_string(char *heading, char *explain, QString Procinfo::*member)
        : Category(heading, explain), str_member(member)
{}

QString Cat_string::string(Procinfo *p)
{
    return p->*str_member;
}

Cat_user::Cat_user(char *heading, char *explain)
        : Cat_string(heading, explain)
{}

QString Cat_user::string(Procinfo *p)
{
    if(p->uid == p->euid)
	return Uidstr::name(p->uid);
    else {
	QString s = Uidstr::name(p->uid);
	s.detach();
	s.append(p->euid == 0 ? "*" : "+");
	return s;
    }
}

Cat_wchan::Cat_wchan(char *heading, char *explain)
        : Cat_string(heading, explain)
{}

QString Cat_wchan::string(Procinfo *p)
{
    return Wchan::name(p->wchan);
}

Cat_cmdline::Cat_cmdline(char *heading, char *explain)
            : Cat_string(heading, explain)
{}

QString Cat_cmdline::string(Procinfo *p)
{
    if(p->cmdline.isEmpty()) {
	QString s(p->comm.length() + 2);
	s[0] = '('; s[1] = '\0';
	s.append(p->comm);
	s.append(")");
	return s;
    } else {
	if(Qps::show_cmd_path)
	    return p->cmdline;
	else {
	    QString s(p->cmdline);
	    s.detach();
	    int i = s.find(' ');
	    if(i < 0)
		i = s.length();
	    if(i > 0) {
		i = s.findRev('/', i - 1);
		if(i >= 0)
		    s.remove(0, i + 1);
	    }
	    return s;
	}
    }
}

Cat_dir::Cat_dir(char *heading, char *explain, char *dirname,
		 QString Procinfo::*member)
    : Cat_string(heading, explain),
      dir(dirname),
      cache(member)
{}

QString Cat_dir::string(Procinfo *p)
{
    if((p->*cache).isNull()) {
	char path[128], buf[512];
	sprintf(path, "%s/%d/%s", procdir, p->pid, (const char *)dir);
	int n = readlink(path, buf, sizeof(buf) - 1);
	if(n < 0) {
	    // Either a kernel process, or access denied.
	    // A hyphen is visually least disturbing here.
	    p->*cache = "-";
	} else {
	    buf[n] = '\0';	// readlink() doesn't append a NUL
	    if(buf[0] == '[') {
		// Linux 2.0 links in [device]:inode form. To resolve it,
		// we just chdir() to it and see where we end up. Perhaps we
		// should change back later? Not really important.
		if(chdir(path) < 0) {
		    p->*cache = "-";
		} else {
		    // getcwd() isn't exactly cheap, but only an issue for
		    // Linux 2.0 anyway.
		    if(!getcwd(buf, sizeof(buf))) {
			p->*cache = "(deleted)";
		    } else
			p->*cache = buf;
		}
	    } else
		p->*cache = buf; // linux >= 2.1.x: path name in link
	}
    }
    return p->*cache;
}

Cat_state::Cat_state(char *heading, char *explain)
        : Category(heading, explain)
{}

QString Cat_state::string(Procinfo *p)
{
    QString s("   ");
    s[0] = p->state;
    s[1] = (p->rss == 0 && p->state != 'Z') ? 'W' : ' ';
    s[2] = (p->nice > 0) ? 'N' : ((p->nice < 0) ? '<' : ' ');
    return s;
}

Cat_policy::Cat_policy(char *heading, char *explain)
           : Category(heading, explain)
{}

QString Cat_policy::string(Procinfo *p)
{
    char c;
    switch(p->get_policy()) {
    case SCHED_FIFO:
	c = 'F'; break;
    case SCHED_RR:
	c = 'R'; break;
    case SCHED_OTHER:
	c = 'O'; break;
    default:
	c = '?'; break;
    }
    QString s(" ");
    s[0] = c;
    return s;
}

int Cat_policy::compare(Procinfo *a, Procinfo *b)
{
    return b->get_policy() - a->get_policy();
}

Cat_rtprio::Cat_rtprio(char *heading, char *explain)
        : Category(heading, explain)
{}

QString Cat_rtprio::string(Procinfo *p)
{
    QString s(5);
    s.setNum(p->get_rtprio());
    return s;
}

int Cat_rtprio::compare(Procinfo *a, Procinfo *b)
{
    return b->get_rtprio() - a->get_rtprio();
}

Cat_time::Cat_time(char *heading, char *explain)
        : Category(heading, explain)
{}

QString Cat_time::string(Procinfo *p)
{
    QString s(8);
    int ticks = p->utime + p->stime;
    if(Qps::cumulative)
	ticks += p->cutime + p->cstime;
    int t = ticks / HZ;		// seconds
    if(t < 10) {
	int hundreds = ticks / (HZ / 100) % 100;
	s.sprintf("%1d.%02ds", t, hundreds);
    } else if(t < 100 * 60) {
	s.sprintf("%2d:%02d", t / 60, t % 60);
    } else if(t < 100 * 3600) {
	int h = t / 3600;
	t %= 3600;
	s.sprintf("%2d:%02dh", h, t / 60);
    } else {
	int d = t / 86400;
	t %= 86400;
	s.sprintf("%dd%dh", d, t / 3600);
    }
    return s;
}

int Cat_time::compare(Procinfo *a, Procinfo *b)
{
    int at = a->utime + a->stime, bt = b->utime + b->stime;
    if(Qps::cumulative) {
	at += a->cutime + a->cstime;
	bt += b->cutime + b->cstime;
    }
    return bt - at;
}

Cat_start::Cat_start(char *heading, char *explain)
          : Category(heading, explain)
{}

QString Cat_start::string(Procinfo *p)
{
    time_t start = p->boot_time + p->starttime / (unsigned)HZ;
    QString s;
    char *ct = ctime(&start);
    if(p->tv.tv_sec - start < 86400) {
	ct[16] = '\0';
	s = ct + 11;
    } else {
	ct[10] = '\0';
	s = ct + 4;
    }
    return s;
}

int Cat_start::compare(Procinfo *a, Procinfo *b)
{
    return b->starttime - a->starttime;
}

Cat_percent::Cat_percent(char *heading, char *explain, int w,
			 float Procinfo::*member)
        : Category(heading, explain), float_member(member), field_width(w)
{}

QString Cat_percent::string(Procinfo *p)
{
    QString s(8);
    s.sprintf("%01.2f", (double)(p->*float_member));
    return s;
}

int Cat_percent::compare(Procinfo *a, Procinfo *b)
{
    float at = a->*float_member, bt = b->*float_member;
    return at < bt ? 1 : (at > bt ? -1 : 0);
}

Cat_tty::Cat_tty(char *heading, char *explain)
        : Category(heading, explain)
{}

QString Cat_tty::string(Procinfo *p)
{
    return Ttystr::name(p->tty);
}

Proc::Proc()
{
    // Note: When adding/removing/changing the fields, the save file
    // version must be increased!
    cats.set(F_PID, new Cat_int("PID", "Process ID", 44, &Procinfo::pid));
    cats.set(F_PPID, new Cat_int("PPID", "Parent process ID", 44,
				 &Procinfo::ppid));
    cats.set(F_PGID, new Cat_int("PGID", "Process group ID", 44,
				 &Procinfo::pgrp));
    cats.set(F_SID, new Cat_int("SID", "Session ID", 44,
				&Procinfo::session));
    cats.set(F_TTY, new Cat_tty("TTY", "Controlling tty"));
    cats.set(F_TPGID, new Cat_int("TPGID", "Process group ID of tty owner", 44,
				  &Procinfo::tpgid));
    cats.set(F_USER, new Cat_user("USER",
				  "Owner (*=suid root, +=suid other user)"));
    cats.set(F_NICE, new Cat_int("NICE",
				"Priority (more positive means less cpu time)",
				 36, &Procinfo::nice));
    cats.set(F_PRI, new Cat_int("PRI",
				"Time left of possible timeslice (rescaled)",
				32, &Procinfo::priority));
    cats.set(F_PLCY, new Cat_policy("PLCY",
				    "Scheduling policy (FIFO, RR or OTHER)"));
    cats.set(F_RPRI, new Cat_rtprio("RPRI",
				  "Realtime priority (0-99, more is better)"));
    cats.set(F_MAJFLT, new Cat_uintl("MAJFLT",
				     "Number of major faults (disk access)",
				     56, &Procinfo::majflt));
    cats.set(F_MINFLT, new Cat_uintl("MINFLT",
				     "Number of minor faults (no disk access)",
				     52, &Procinfo::minflt));
    cats.set(F_TRS, new Cat_uintl("TRS", "Text resident set size in Kbytes",
				  40, &Procinfo::trs));
    cats.set(F_DRS, new Cat_uintl("DRS", "Data resident set size in Kbytes",
				  40, &Procinfo::drs));
    cats.set(F_SIZE, new Cat_uintl("SIZE",
				   "Virtual image size of process in Kbytes",
				   40, &Procinfo::size));
    cats.set(F_SWAP, new Cat_swap("SWAP", "Kbytes on swap device"));
    cats.set(F_RSS, new Cat_uintl("RSS",
				  "Resident set size; Kbytes of program "
				  "in memory", 40, &Procinfo::resident));
    cats.set(F_SHARE, new Cat_uintl("SHARE", "Shared memory in Kbytes",
				    48, &Procinfo::share));
    cats.set(F_DT, new Cat_uintl("DT", "Number of dirty (non-written) pages",
				 40, &Procinfo::dt));
    cats.set(F_STAT, new Cat_state("STAT", "State of the process"));
    cats.set(F_FLAGS, new Cat_hex("FLAGS", "Process flags (hex)", 60,
				  &Procinfo::flags));
    cats.set(F_WCHAN, new Cat_wchan("WCHAN",
				 "Kernel function where process is sleeping"));
    cats.set(F_UID, new Cat_int("UID", "User ID", 40, &Procinfo::uid));
    cats.set(F_WCPU, new Cat_percent("%WCPU",
				   "Weighted percentage of CPU (30 s average)",
				     52, &Procinfo::wcpu));
    cats.set(F_CPU, new Cat_percent("%CPU",
				    "Percentage of CPU used since last update",
				    42, &Procinfo::pcpu));
    cats.set(F_MEM, new Cat_percent("%MEM",
				  "Percentage of memory used (RSS/total mem)",
				    46, &Procinfo::pmem));
    cats.set(F_START, new Cat_start("START", "Time process started"));
    cats.set(F_TIME, new Cat_time("TIME", "Total CPU time used since start"));
    cats.set(F_COMM, new Cat_string("COMM", "Command that started the process",
				    &Procinfo::comm));
    cats.set(F_CWD, new Cat_dir("CWD", "Current working directory",
				"cwd", &Procinfo::cwd));
    cats.set(F_ROOT, new Cat_dir("ROOT", "Root directory of process",
				"root", &Procinfo::root));
    cats.set(F_CMDLINE, new Cat_cmdline("CMDLINE",
				     "Command line that started the process"));

    for(int i = 0; i < cats.size(); i++)
	cats[i]->index = i;

    Procinfo::init_static();

#if 0
    // try downcasing the headings, to see if that's more readable
    // (but less ps/top compatible, probably not worth it)
    for(int i = 0; i < cats.size(); i++) {
	char *p = strdup(cats[i]->name);
	for(char *q = p + 1 + (*p == '%'); *q; q++)
	    *q += 'a' - 'A';
	cats[i]->name = p;
    }
#endif
}

// update the process list
void Proc::refresh()
{
    Procinfo::read_common();
    Svec<Procinfo *> oldprocs = procs;
    procs.clear();
    DIR *d = opendir(procdir);
    struct dirent *e;
    while((e = readdir(d)) != 0) {
	if(e->d_name[0] >= '0' && e->d_name[0] <= '9') {
	    Procinfo *pi = new Procinfo(atoi(e->d_name));
	    if(pi->pid == -1)
		delete pi;		// already gone
	    else {
		procs.add(pi);
	    }
	}
    }
    closedir(d);
    // correlate information from old generation
    for(int i = 0; i < procs.size(); i++) {
	Procinfo *ip = procs[i];
	int j, pid = ip->pid;
	if(i < oldprocs.size() && pid == oldprocs[i]->pid)
	    j = i;	 // heuristic to avoid quadratic behaviour
	else {
	    for(j = 0; j < oldprocs.size(); j++)
		if(pid == oldprocs[j]->pid)
		    break;
	}
	if(j < oldprocs.size()) {
	    Procinfo *jp = oldprocs[j];
	    
	    // calculate pcpu and wcpu from previous procinfo
	    int dt = (ip->tv.tv_usec - jp->tv.tv_usec) / (1000000 / HZ)
		     + (ip->tv.tv_sec - jp->tv.tv_sec) * HZ;
	    int dcpu = ip->stime - jp->stime + ip->utime - jp->utime;
	    ip->pcpu = 100.0 * dcpu / dt;
	    if(ip->pcpu > 99.99) ip->pcpu = 99.99;

	    const float a = Procview::avg_factor;
	    ip->wcpu = a * jp->wcpu + (1 - a) * ip->pcpu;

	    // propagate some fields to new incarnation
	    ip->selected = jp->selected;
	    ip->details = jp->details;
	    jp->details = 0;
	    if(ip->details)
		ip->details->set_procinfo(ip);
	} else {
	    // %cpu first time = (cpu time since start) / (time since start)
	    int jiffies_since_boot = ip->tv.tv_usec / (1000000 / HZ)
		                   + (ip->tv.tv_sec - ip->boot_time) * HZ;
	    int dt = jiffies_since_boot - ip->starttime;
	    int dcpu = ip->stime + ip->utime;
	    ip->pcpu = 100.0 * dcpu / dt;
	    if(dt == 0 || ip->pcpu > 99.99 || ip->pcpu < 0)
		ip->pcpu = 0.0;
	    ip->selected = FALSE;
	    ip->wcpu = ip->pcpu;	// just a start
	}
    }
    // remove useless old Procinfos
    for(int i = 0; i < oldprocs.size(); i++) {
	Procinfo *oi = oldprocs[i];
	if(oi->details)
	    oi->details->process_gone();
	if(!oi->viewed) {
	    delete oi;
	}
    }
}

int Procview::user_fields[] = {F_PID, F_TTY, F_USER, F_NICE, F_SIZE, F_RSS,
			       F_STAT, F_CPU, F_START, F_TIME,
			       F_CMDLINE, F_END};
int Procview::jobs_fields[] = {F_PID, F_PPID, F_PGID, F_SID, F_TTY, F_TPGID,
			       F_STAT, F_UID, F_TIME, F_CMDLINE, F_END};
int Procview::mem_fields[] = {F_PID, F_TTY, F_MAJFLT, F_MINFLT, F_TRS, F_DRS,
			      F_SIZE, F_SWAP, F_RSS, F_SHARE, F_DT, F_CMDLINE,
			      F_END};

float Procview::avg_factor = 1.0;

Procview::Procview(Proc *p)
        : proc(p)
{
    procs = p->procs;
    sortcat = p->cats[F_WCPU];
    reversed = FALSE;
    viewproc = ALL;
    viewfields = USER;
    set_fields();
}

// read new process info
void Procview::refresh()
{
    old_procs.purge();
    old_procs = procs;
    procs.clear();
    proc->refresh();
    rebuild();
}

// re-sort the process info
void Procview::rebuild()
{
    int my_uid;
    switch(viewproc) {
    case ALL:
	procs = proc->procs;
	break;
    case OWNED:
	my_uid = getuid();
	procs.clear();
	for(int i = 0; i < proc->procs.size(); i++) {
	    Procinfo *p = proc->procs[i];
	    if(my_uid == p->uid)
		procs.add(p);
	    else
		p->selected = 0;
	}
	break;
    case NROOT:
	procs.clear();
	for(int i = 0; i < proc->procs.size(); i++) {
	    Procinfo *p = proc->procs[i];
	    if(p->uid != 0)
		procs.add(p);
	    else
		p->selected = 0;
	}
	break;
    case RUNNING:
	procs.clear();
	for(int i = 0; i < proc->procs.size(); i++) {
	    Procinfo *p = proc->procs[i];
	    if(strchr("RDW", p->state) != 0)
		procs.add(p);
	    else
		p->selected = 0;
	}
	break;
    }
    static_sortcat = sortcat;
    procs.sort(reversed ? compare_backwards : compare);
    for(int i = 0; i < proc->procs.size(); i++)
	proc->procs[i]->viewed = FALSE;
    for(int i = 0; i < procs.size(); i++)
	procs[i]->viewed = TRUE;
}

void Procview::set_fields_list(int fields[])
{
    cats.clear();
    for(int i = 0; fields[i] != F_END; i++)
	cats.add(proc->cats[fields[i]]);
}

void Procview::set_fields()
{
    switch(viewfields) {
    case USER:
	set_fields_list(user_fields);
	break;
    case JOBS:
	set_fields_list(jobs_fields);
	break;
    case MEM:
	set_fields_list(mem_fields);
	break;
    case CUSTOM:
	;
    }
}

// return the column number of a field, or -1 if not displayed
int Procview::findCol(int field)
{
    for(int i = 0; i < cats.size(); i++)
	if(cats[i] == proc->cats[field])
	    return i;
    return -1;
}

// add a category at the right place (no duplication check)
void Procview::add_cat(Category *c)
{
    int i = cats.size() - 1;
    cats.setSize(i + 2);
    while(i >= 0 && cats[i]->index > c->index) {
	cats[i + 1] = cats[i];
	i--;
    }
    cats[i + 1] = c;
}

void Procview::remove_cat(int index)
{
    cats.remove(index);
}

Category *Procview::static_sortcat = 0;

int Procview::compare(Procinfo **a, Procinfo **b)
{
    int r = static_sortcat->compare(*a, *b);
    return (r == 0) ? ((*a)->pid > (*b)->pid ? 1 : -1) : r;
}

int Procview::compare_backwards(Procinfo **a, Procinfo **b)
{
    int r = static_sortcat->compare(*a, *b);
    return (r == 0) ? ((*a)->pid > (*b)->pid ? 1 : -1) : -r;
}

